# vi: ts=4 sw=4 sts=4 expandtab
'''
:mod:`ophyd.utils` - Miscellaneous utility functions
====================================================

.. module:: ophyd.utils
   :synopsis:
'''

import logging
from collections import OrderedDict

from .errors import *
from .epics_pvs import *
from .paths import makedirs, make_dir_tree


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def enum(**enums):
    '''Create an enum from the keyword arguments'''
    return type('Enum', (object,), enums)


class OrderedDefaultDict(OrderedDict):
    """
    a combination of defaultdict and OrderedDict

    source: http://stackoverflow.com/a/6190500/1221924
    """
    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None and not callable(default_factory)):
            raise TypeError('first argument must be callable')
        super().__init__(*a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, self.items()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        import copy
        return type(self)(self.default_factory,
                          copy.deepcopy(self.items()))

    def __repr__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, self.default_factory,
                               super().__repr__())



import sys
import logging
import unittest
import threading
import random
import time
import copy
import pytest
from functools import wraps
import weakref

import numpy as np
import epics

from ophyd.signal import (Signal, EpicsSignal, EpicsSignalRO, DerivedSignal)
from ophyd.utils import ReadOnlyError
from ophyd.status import wait

logger = logging.getLogger(__name__)


_FAKE_PV_LIST = []


class FakeEpicsPV(object):
    _connect_delay = (0.05, 0.1)
    _update_rate = 0.1
    fake_values = (0.1, 0.2, 0.3)
    _pv_idx = 0
    auto_monitor = True

    def __init__(self, pvname, form=None,
                 callback=None, connection_callback=None,
                 auto_monitor=True, enum_strs=None,
                 **kwargs):

        global _FAKE_PV_LIST
        _FAKE_PV_LIST.append(self)

        self._pvname = pvname
        self._connection_callback = connection_callback
        self._form = form
        self._auto_monitor = auto_monitor
        self._value = self.fake_values[0]
        self._connected = False
        self._running = True
        self.enum_strs = enum_strs
        FakeEpicsPV._pv_idx += 1
        self._idx = FakeEpicsPV._pv_idx

        self._update = True

        self._lock = threading.Lock()
        self._thread = threading.Thread(target=self._update_loop)
        self._thread.daemon = True
        self._thread.start()

        # callbacks mechanism copied from pyepics
        # ... but tweaked with a weakvaluedictionary so PV objects get
        # destructed
        self.callbacks = weakref.WeakValueDictionary()

        if callback:
            self.add_callback(callback)

    def __del__(self):
        self.clear_callbacks()
        self._running = False

        try:
            self._thread.join()
            self._thread = None
        except Exception:
            pass

    def get_timevars(self):
        pass

    def get_ctrlvars(self):
        pass

    @property
    def connected(self):
        return self._connected

    def wait_for_connection(self, timeout=None):
        if self._pvname in ('does_not_connect', ):
            return False

        while not self._connected:
            time.sleep(0.05)

        return True

    def _update_loop(self):
        time.sleep(random.uniform(*self._connect_delay))
        if self._connection_callback is not None:
            self._connection_callback(pvname=self._pvname, conn=True, pv=self)

        if self._pvname in ('does_not_connect', ):
            return

        self._connected = True
        last_value = None

        while self._running:
            with self._lock:
                if self._update:
                    self._value = random.choice(self.fake_values)

                if self._value != last_value:
                    sys.stdout.flush()
                    self.run_callbacks()
                    last_value = self._value

                time.sleep(self._update_rate)

            time.sleep(0.01)

    @property
    def lower_ctrl_limit(self):
        return min(self.fake_values)

    @property
    def upper_ctrl_limit(self):
        return max(self.fake_values)

    def run_callbacks(self):
        for index in sorted(list(self.callbacks.keys())):
            if not self._running:
                break
            self.run_callback(index)

    def run_callback(self, index):
        fcn = self.callbacks[index]
        kwd = dict(pvname=self._pvname,
                   count=1,
                   nelm=1,
                   type=None,
                   typefull=None,
                   ftype=None,
                   access='rw',
                   chid=self._idx,
                   read_access=True,
                   write_access=True,
                   value=self.value,
                   )

        kwd['cb_info'] = (index, self)
        if hasattr(fcn, '__call__'):
            fcn(**kwd)

    def add_callback(self, callback=None, index=None, run_now=False,
                     with_ctrlvars=True):
        if hasattr(callback, '__call__'):
            if index is None:
                index = 1
                if len(self.callbacks) > 0:
                    index = 1 + max(self.callbacks.keys())
            self.callbacks[index] = callback

        if run_now:
            if self.connected:
                self.run_callback(index)
        return index

    def remove_callback(self, index=None):
        if index in self.callbacks:
            self.callbacks.pop(index)

    def clear_callbacks(self):
        self.callbacks.clear()

    @property
    def precision(self):
        return 0

    @property
    def units(self):
        return str(None)

    @property
    def timestamp(self):
        return time.time()

    @property
    def pvname(self):
        return self._pvname

    @property
    def value(self):
        return self._value

    def __repr__(self):
        return '<FakePV %s value=%s>' % (self._pvname, self.value)

    def get(self, as_string=False, use_numpy=False,
            use_monitor=False):
        if as_string:

            if isinstance(self.value, list):
                if self.enum_strs:
                    return [self.enum_strs[_] for _ in self.value]
                return list(self.value)
            if isinstance(self.value, str):
                return self.value
            else:
                if self.enum_strs:
                    return self.enum_strs[self.value]
                return str(self.value)
        elif use_numpy:
            return np.array(self.value)
        else:
            return self.value

    def put(self, value, wait=False, timeout=30.0,
            use_complete=False, callback=None, callback_data=None):

        with self._lock:
            self._update = False
            self._value = value


class FakeEpicsWaveform(FakeEpicsPV):
    strings = ['abcd', 'efgh', 'ijkl']
    fake_values = [[ord(c) for c in s] + [0]
                   for s in strings]
    auto_monitor = False
    form = 'time'


def _cleanup_fake_pvs():
    pvs = list(_FAKE_PV_LIST)
    del _FAKE_PV_LIST[:]

    for pv in pvs:
        pv.clear_callbacks()
        pv._running = False
        pv._connection_callback = None

    for pv in pvs:
        try:
            pv._thread.join()
            pv._thread = None
        except Exception:
            pass


def using_fake_epics_pv(fcn):
    @wraps(fcn)
    def wrapped(*args, **kwargs):
        pv_backup = epics.PV
        epics.PV = FakeEpicsPV
        try:
            return fcn(*args, **kwargs)
        finally:
            epics.PV = pv_backup
            _cleanup_fake_pvs()

    return wrapped


def using_fake_epics_waveform(fcn):
    @wraps(fcn)
    def wrapped(*args, **kwargs):
        pv_backup = epics.PV
        epics.PV = FakeEpicsWaveform
        try:
            return fcn(*args, **kwargs)
        finally:
            epics.PV = pv_backup
            _cleanup_fake_pvs()

    return wrapped


@using_fake_epics_pv
def test_fakepv():
    pvname = 'fakepv_nowaythisexists' * 10

    info = dict(called=False)

    def conn(**kwargs):
        info['conn'] = True
        info['conn_kw'] = kwargs

    def value_cb(**kwargs):
        info['value'] = True
        info['value_kw'] = kwargs

    pv = epics.PV(pvname, callback=value_cb, connection_callback=conn,
                  )

    if not pv.wait_for_connection():
        raise ValueError('should return True on connection')

    assert pv.pvname == pvname

    pv._update_rate = 0.5
    time.sleep(0.2)

    assert info['conn']
    assert info['value']
    assert info['value_kw']['value'] == pv.value


def test_signal_base():
    start_t = time.time()

    name = 'test'
    value = 10.0
    signal = Signal(name=name, value=value, timestamp=start_t)
    signal.wait_for_connection()

    assert signal.connected
    assert signal.name == name
    assert signal.value == value
    assert signal.get() == value
    assert signal.timestamp == start_t

    info = dict(called=False)

    def _sub_test(**kwargs):
        info['called'] = True
        info['kw'] = kwargs

    signal.subscribe(_sub_test, run=False)
    assert not info['called']

    signal.value = value
    signal.clear_sub(_sub_test)

    signal.subscribe(_sub_test, run=False)
    signal.clear_sub(_sub_test, event_type=signal.SUB_VALUE)

    kw = info['kw']
    assert 'value' in kw
    assert 'timestamp' in kw
    assert 'old_value' in kw

    assert kw['value'] == value
    assert kw['old_value'] == value
    assert kw['timestamp'] == signal.timestamp

    # readback callback for soft signal
    info = dict(called=False)
    signal.subscribe(_sub_test, event_type=Signal.SUB_VALUE,
                     run=False)
    assert not info['called']
    signal.put(value + 1)
    assert info['called']

    signal.clear_sub(_sub_test)
    kw = info['kw']

    assert 'value' in kw
    assert 'timestamp' in kw
    assert 'old_value' in kw

    assert kw['value'] == value + 1
    assert kw['old_value'] == value
    assert kw['timestamp'] == signal.timestamp

    signal.trigger()
    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))


def test_signal_copy():
    start_t = time.time()

    name = 'test'
    value = 10.0
    signal = Signal(name=name, value=value, timestamp=start_t)
    sig_copy = copy.copy(signal)

    assert signal.name == sig_copy.name
    assert signal.value == sig_copy.value
    assert signal.get() == sig_copy.get()
    assert signal.timestamp == sig_copy.timestamp


def test_rw_removal():
    # rw kwarg is no longer used
    with pytest.raises(RuntimeError):
        EpicsSignal('readpv', rw=False)

    with pytest.raises(RuntimeError):
        EpicsSignal('readpv', rw=True)


@using_fake_epics_pv
def test_epicssignal_readonly():
    signal = EpicsSignalRO('readpv')
    signal.wait_for_connection()
    signal.value

    with pytest.raises(ReadOnlyError):
        signal.value = 10

    with pytest.raises(ReadOnlyError):
        signal.put(10)

    with pytest.raises(ReadOnlyError):
        signal.set(10)

    # vestigial, to be removed
    with pytest.raises(AttributeError):
        signal.setpoint_ts

    # vestigial, to be removed
    with pytest.raises(AttributeError):
        signal.setpoint

    signal.precision
    signal.timestamp
    signal.limits

    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))
    time.sleep(0.2)


@using_fake_epics_pv
def test_epicssignal_readwrite_limits():
    signal = EpicsSignal('readpv', write_pv='readpv', limits=True)

    signal.wait_for_connection()
    signal.check_value((signal.low_limit + signal.high_limit) / 2)

    try:
        signal.check_value(None)
    except ValueError:
        pass
    else:
        raise ValueError('value=None')

    try:
        signal.check_value(signal.low_limit - 1)
    except ValueError:
        pass
    else:
        raise ValueError('lower limit %s' % (signal.limits, ))

    try:
        signal.check_value(signal.high_limit + 1)
    except ValueError:
        pass
    else:
        raise ValueError('upper limit')


@using_fake_epics_pv
def test_epicssignal_readwrite():
    signal = EpicsSignal('readpv', write_pv='writepv')

    signal.wait_for_connection()
    assert signal.setpoint_pvname == 'writepv'
    assert signal.pvname == 'readpv'
    signal.value

    signal._update_rate = 2
    time.sleep(0.2)

    value = 10
    signal.value = value
    signal.setpoint = value
    assert signal.setpoint == value
    signal.setpoint_ts

    signal.limits
    signal.precision
    signal.timestamp

    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))
    time.sleep(0.2)


@using_fake_epics_waveform
def test_epicssignal_waveform():
    def update_cb(value=None, **kwargs):
        assert value in FakeEpicsWaveform.strings

    signal = EpicsSignal('readpv', string=True)

    signal.wait_for_connection()

    signal.subscribe(update_cb)
    assert signal.value in FakeEpicsWaveform.strings


@using_fake_epics_pv
def test_no_connection():
    # special case in FakeEpicsPV that returns false in wait_for_connection
    sig = EpicsSignal('does_not_connect')
    pytest.raises(TimeoutError, sig.wait_for_connection)

    sig = EpicsSignal('does_not_connect')
    pytest.raises(TimeoutError, sig.put, 0.0)
    pytest.raises(TimeoutError, sig.get)

    sig = EpicsSignal('connects', write_pv='does_not_connect')
    pytest.raises(TimeoutError, sig.wait_for_connection)


@using_fake_epics_pv
def test_enum_strs():
    sig = EpicsSignal('connects')
    sig.wait_for_connection()

    enums = ['enum_strs']

    # hack this onto the FakeEpicsPV
    sig._read_pv.enum_strs = enums

    assert sig.enum_strs == enums


@using_fake_epics_pv
def test_setpoint():
    sig = EpicsSignal('connects')
    sig.wait_for_connection()

    sig.get_setpoint()
    sig.get_setpoint(as_string=True)


def test_epicssignalro():
    # not in initializer parameters anymore
    pytest.raises(TypeError, EpicsSignalRO, 'test',
                      write_pv='nope_sorry')

@using_fake_epics_pv
def test_describe():
    sig = EpicsSignal('my_pv')
    sig._write_pv.enum_strs = ('enum1', 'enum2')
    sig.wait_for_connection()

    sig.put(1)
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'integer'
    assert desc['shape'] == []
    assert 'precision' in desc
    assert 'enum_strs' in desc
    assert 'upper_ctrl_limit' in desc
    assert 'lower_ctrl_limit' in desc

    sig.put('foo')
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'string'
    assert desc['shape'] == []

    sig.put(3.14)
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'number'
    assert desc['shape'] == []

    import numpy as np
    sig.put(np.array([1,]))
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'array'
    assert desc['shape'] == [1,]


def test_set_method():
    sig = Signal()

    st = sig.set(28)
    wait(st)
    assert st.done
    assert st.success
    assert sig.get() == 28


def test_soft_derived():
    timestamp = 1.0
    value = 'q'
    original = Signal(name='original', timestamp=timestamp, value=value)

    cb_values = []

    def callback(value=None, **kwargs):
        nonlocal cb_values
        cb_values.append(value)

    derived = DerivedSignal(derived_from=original, name='derived')
    derived.subscribe(callback)

    assert derived.timestamp == timestamp
    assert derived.get() == value
    assert derived.timestamp == timestamp
    assert derived.describe()[derived.name]['derived_from'] == original.name

    new_value = 'r'
    derived.put(new_value)
    assert original.get() == new_value
    assert derived.get() == new_value
    assert derived.timestamp == original.timestamp
    assert derived.limits == original.limits

    copied = copy.copy(derived)
    assert copied.derived_from.value == original.value
    assert copied.derived_from.timestamp == original.timestamp
    assert copied.derived_from.name == original.name

    derived.put('s')
    assert cb_values == ['r', 's']


@using_fake_epics_pv
def test_epics_signal_derived():
    signal = EpicsSignalRO('fakepv', name='original')

    derived = DerivedSignal(derived_from=signal, name='derived')
    derived.wait_for_connection()

    derived.connected

    # race condition with the FakeEpicsPV update loop, can't really test
    # assert derived.timestamp == signal.timestamp
    # assert derived.get() == signal.value


@pytest.mark.parametrize('put_complete', [True, False])
def test_epicssignal_set(put_complete):
    sim_pv = EpicsSignal(write_pv='XF:31IDA-OP{Tbl-Ax:X1}Mtr.VAL',
                         read_pv='XF:31IDA-OP{Tbl-Ax:X1}Mtr.RBV',
                         put_complete=put_complete)
    sim_pv.wait_for_connection()

    logging.getLogger('ophyd.signal').setLevel(logging.DEBUG)
    logging.getLogger('ophyd.utils.epics_pvs').setLevel(logging.DEBUG)
    print('tolerance=', sim_pv.tolerance)
    assert sim_pv.tolerance is not None

    start_pos = sim_pv.get()

    # move to +0.2 and check the status object
    target = sim_pv.get() + 0.2
    st = sim_pv.set(target, timeout=1, settle_time=0.001)
    wait(st)
    assert st.done
    assert st.success
    print('status 1', st)
    assert abs(target - sim_pv.get()) < 0.05

    # move back to -0.2, forcing a timeout with a low value
    target = sim_pv.get() - 0.2
    st = sim_pv.set(target, timeout=1e-6)
    time.sleep(0.1)
    print('status 2', st)
    assert st.done
    assert not st.success

    # keep the axis in position
    st = sim_pv.set(start_pos)
    wait(st)


def test_epicssignal_alarm_status():
    sig = EpicsSignal(write_pv='XF:31IDA-OP{Tbl-Ax:X1}Mtr.VAL',
                      read_pv='XF:31IDA-OP{Tbl-Ax:X1}Mtr.RBV')
    sig.alarm_status
    sig.alarm_severity
    sig.setpoint_alarm_status
    sig.setpoint_alarm_severity


def test_epicssignalro_alarm_status():
    sig = EpicsSignalRO('XF:31IDA-OP{Tbl-Ax:X1}Mtr.RBV')
    sig.alarm_status
    sig.alarm_severity


from . import main
is_main = (__name__ == '__main__')
main(is_main)



import time
from threading import RLock
from functools import wraps

import logging
import threading
import numpy as np

logger = logging.getLogger(__name__)


# This is used below by StatusBase.
def _locked(func):
    "an decorator for running a method with the instance's lock"
    @wraps(func)
    def f(self, *args, **kwargs):
        with self._lock:
            func(self, *args, **kwargs)

    return f


class StatusBase:
    """
    This is a base class that provides a single-slot callback for when the
    specific operation has finished.

    Parameters
    ----------
    timeout : float, optional
        The default timeout to use for a blocking wait, and the amount of time
        to wait to mark the operation as failed
    settle_time : float, optional
        The amount of time to wait between the caller specifying that the
        status has completed to running callbacks
    """
    def __init__(self, *, timeout=None, settle_time=None, done=False,
                 success=False):
        super().__init__()
        self._lock = RLock()
        self._cb = None
        self.done = done
        self.success = success
        self.timeout = None

        if settle_time is None:
            settle_time = 0.0

        self.settle_time = float(settle_time)

        if timeout is not None:
            self.timeout = float(timeout)

        if self.done:
            # in the case of a pre-completed status object,
            # don't handle timeout
            return

        if self.timeout is not None and self.timeout > 0.0:
            thread = threading.Thread(target=self._timeout_thread, daemon=True)
            self._timeout_thread = thread
            self._timeout_thread.start()

    def _timeout_thread(self):
        '''Handle timeout'''
        try:
            wait(self, timeout=self.timeout + self.settle_time,
                 poll_rate=max(1.0, self.timeout / 10.0))
        except TimeoutError:
            logger.debug('Status object %s timed out', str(self))
            try:
                self._handle_failure()
            finally:
                self._finished(success=False)
        except RuntimeError:
            pass
        finally:
            self._timeout_thread = None

    def _handle_failure(self):
        pass

    def _settled(self):
        '''Hook for when status has completed and settled'''
        pass

    def _settle_then_run_callbacks(self, success=True):
        # wait until the settling time is done to mark completion
        if self.settle_time > 0.0:
            time.sleep(self.settle_time)

        with self._lock:
            self.success = success
            self.done = True
            self._settled()

            if self._cb is not None:
                self._cb()
                self._cb = None

    def _finished(self, success=True, **kwargs):
        '''Inform the status object that it is done and if it succeeded

        .. warning::

           kwargs are not used, but are accepted because pyepics gives
           in a bunch of kwargs that we don't care about.  This allows
           the status object to be handed directly to pyepics (but
           this is probably a bad idea for other reason.

           This may be deprecated in the future.

        Parameters
        ----------
        success : bool, optional
           if the action succeeded.
        '''
        if self.done:
            return

        if success and self.settle_time > 0:
            # delay gratification until the settle time is up
            self._settle_thread = threading.Thread(
                target=self._settle_then_run_callbacks, daemon=True,
                kwargs=dict(success=success),
            )
            self._settle_thread.start()
        else:
            self._settle_then_run_callbacks(success=success)

    @property
    def finished_cb(self):
        """
        Callback to be run when the status is marked as finished

        The callback has no arguments ::

            def cb() -> None:

        """
        return self._cb

    @finished_cb.setter
    @_locked
    def finished_cb(self, cb):
        if self._cb is not None:
            raise RuntimeError("Can not change the call back")
        if self.done:
            cb()
        else:
            self._cb = cb

    def __str__(self):
        return ('{0}(done={1.done}, '
                'success={1.success})'
                ''.format(self.__class__.__name__, self)
                )

    __repr__ = __str__


class Status(StatusBase):
    '''A basic status object

    Has an optional associated object instance

    Attributes
    ----------
    obj : any or None
        The object
    '''
    def __init__(self, obj=None, **kwargs):
        self.obj = obj
        super().__init__(**kwargs)

    def __str__(self):
        return ('{0}(obj={1.obj}, '
                'done={1.done}, '
                'success={1.success})'
                ''.format(self.__class__.__name__, self)
                )

    __repr__ = __str__


class DeviceStatus(StatusBase):
    '''Device status

    Parameters
    ----------
    device : obj
    done : bool, optional
        Whether or not the motion has already completed
    success : bool, optional
        If motion has already completed, the status of that motion
    timeout : float, optional
        The default timeout to use for a blocking wait, and the amount of time
        to wait to mark the motion as failed
    settle_time : float, optional
        The amount of time to wait between motion completion and running
        callbacks
    '''
    def __init__(self, device, **kwargs):
        self.device = device
        super().__init__(**kwargs)

    def _handle_failure(self):
        super()._handle_failure()
        logger.debug('Trying to stop %s', str(self.device))
        self.device.stop()

    def __str__(self):
        return ('{0}(device={1.device.name}, done={1.done}, '
                'success={1.success})'
                ''.format(self.__class__.__name__, self)
                )

    __repr__ = __str__


class MoveStatus(DeviceStatus):
    '''Asynchronous movement status

    Parameters
    ----------
    positioner : Positioner
    target : float or array-like
        Target position
    done : bool, optional
        Whether or not the motion has already completed
    success : bool, optional
        If motion has already completed, the status of that motion
    start_ts : float, optional
        The motion start timestamp
    timeout : float, optional
        The default timeout to use for a blocking wait, and the amount of time
        to wait to mark the motion as failed
    settle_time : float, optional
        The amount of time to wait between motion completion and running
        callbacks

    Attributes
    ----------
    pos : Positioner
    target : float or array-like
        Target position
    done : bool
        Whether or not the motion has already completed
    start_ts : float
        The motion start timestamp
    finish_ts : float
        The motion completd timestamp
    finish_pos : float or ndarray
        The final position
    success : bool
        Motion successfully completed
    '''

    def __init__(self, positioner, target, *, start_ts=None,
                 **kwargs):
        if start_ts is None:
            start_ts = time.time()

        self.pos = positioner
        self.target = target
        self.start_ts = start_ts
        self.finish_ts = None
        self.finish_pos = None

        # call the base class
        super().__init__(positioner, **kwargs)

    @property
    def error(self):
        '''Error between target position and current* position

        * If motion is already complete, the final position is used
        '''
        if self.finish_pos is not None:
            finish_pos = self.finish_pos
        else:
            finish_pos = self.pos.position

        try:
            return np.array(finish_pos) - np.array(self.target)
        except Exception:
            return None

    def _settled(self):
        '''Hook for when motion has completed and settled'''
        super()._settled()
        self.finish_ts = time.time()
        self.finish_pos = self.pos.position

    @property
    def elapsed(self):
        '''Elapsed time'''
        if self.finish_ts is None:
            return time.time() - self.start_ts
        else:
            return self.finish_ts - self.start_ts

    def __str__(self):
        return ('{0}(done={1.done}, pos={1.pos.name}, '
                'elapsed={1.elapsed:.1f}, '
                'success={1.success}, settle_time={1.settle_time})'
                ''.format(self.__class__.__name__, self)
                )

    __repr__ = __str__


def wait(status, timeout=None, *, poll_rate=0.05):
    '''(Blocking) wait for the status object to complete

    Parameters
    ----------
    timeout : float, optional
        Amount of time in seconds to wait. None disables, such that wait() will
        only return when either the status completes or if interrupted by the
        user.
    poll_rate : float, optional
        Polling rate used to check the status

    Raises
    ------
    TimeoutError
        If time waited exceeds specified timeout
    RuntimeError
        If the status failed to complete successfully
    '''
    t0 = time.time()

    def time_exceeded():
        return timeout is not None and (time.time() - t0) > timeout

    while not status.done and not time_exceeded():
        time.sleep(poll_rate)

    if status.done:
        if status.success is not None and not status.success:
            raise RuntimeError('Operation completed but reported an error')
    elif time_exceeded():
        elapsed = time.time() - t0
        raise TimeoutError('Operation failed to complete within {} seconds'
                           '(elapsed {} sec)'.format(timeout, elapsed))



import logging
import unittest
import pytest

from unittest.mock import Mock
from ophyd.ophydobj import OphydObject
from ophyd.status import (StatusBase, DeviceStatus, wait)

from . import main

logger = logging.getLogger(__name__)


def test_status_basic():
    st = StatusBase()
    st._finished()


def test_status_callback():
    st = StatusBase()
    cb = Mock()

    st.finished_cb = cb
    assert st.finished_cb is cb
    with pytest.raises(RuntimeError):
        st.finished_cb = None

    st._finished()
    cb.assert_called_once_with()


def test_status_others():
    DeviceStatus(None)


def test_status_wait():
    st = StatusBase()
    st._finished()
    wait(st)


def test_wait_status_failed():
    st = StatusBase(timeout=0.05)
    with pytest.raises(RuntimeError):
        wait(st)


def test_status_wait_timeout():
    st = StatusBase()

    with pytest.raises(TimeoutError):
        wait(st, timeout=0.05)


def test_ophydobj():
    parent = OphydObject(name='name', parent=None)
    child = OphydObject(name='name', parent=parent)
    assert child.parent is parent

    child._run_sub('not_a_callable', sub_type='sub')

    with pytest.raises(ValueError):
        child.subscribe(None, event_type=None)

    with pytest.raises(KeyError):
        child.subscribe(None, event_type='unknown_event_type')

    assert parent.connected


def test_self_removing_cb():
    class TestObj(OphydObject):
        SUB_TEST = 'value'

    test_obj = TestObj(name='name', parent=None)

    hit_A = 0
    hit_B = 0

    def remover(obj, **kwargs):
        nonlocal hit_A
        hit_A += 1
        obj.clear_sub(remover)

    def sitter(**kwargs):
        nonlocal hit_B
        hit_B += 1

    test_obj.subscribe(remover, 'value')
    test_obj.subscribe(sitter, 'value')
    test_obj._run_subs(sub_type='value')

    assert hit_A == 1
    assert hit_B == 1

    test_obj._run_subs(sub_type='value')

    assert hit_A == 1
    assert hit_B == 2


is_main = (__name__ == '__main__')
main(is_main)



''' '''

import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

from .base import *
from .cam import *
from .detectors import *
from .plugins import *
from .trigger_mixins import *




import logging
import pytest
from io import StringIO

from ophyd import (SimDetector, SingleTrigger, Component,
                   DynamicDeviceComponent)
from ophyd.areadetector.plugins import (ImagePlugin, StatsPlugin,
                                        ColorConvPlugin,
                                        ProcessPlugin, OverlayPlugin,
                                        ROIPlugin, TransformPlugin,
                                        NetCDFPlugin, TIFFPlugin, JPEGPlugin,
                                        HDF5Plugin,
                                        MagickPlugin)
# we do not have nexus installed on our test IOC
# from ophyd.areadetector.plugins import NexusPlugin
from ophyd.areadetector.plugins import PluginBase
from ophyd.areadetector.util import stub_templates
from ophyd.device import (Component as Cpt, )

logger = logging.getLogger(__name__)

prefix = 'XF:31IDA-BI{Cam:Tbl}'
ad_path = '/epics/support/areaDetector/1-9-1/ADApp/Db/'


# lifted from soft-matter/pims source
def _recursive_subclasses(cls):
    "Return all subclasses (and their subclasses, etc.)."
    # Source: http://stackoverflow.com/a/3862957/1221924
    return (cls.__subclasses__() +
            [g for s in cls.__subclasses__()
             for g in _recursive_subclasses(s)])


def test_basic():
    class MyDetector(SingleTrigger, SimDetector):
        tiff1 = Cpt(TIFFPlugin, 'TIFF1:')

    det = MyDetector(prefix)
    det.wait_for_connection()
    det.stage()
    det.trigger()
    det.unstage()


def test_stubbing():
    try:
        for line in stub_templates(ad_path):
            logger.debug('Stub line: %s', line)
    except OSError:
        # self.fail('AreaDetector db path needed to run test')
        pass


def test_detector():
    det = SimDetector(prefix)

    det.find_signal('a', f=StringIO())
    det.find_signal('a', use_re=True, f=StringIO())
    det.find_signal('a', case_sensitive=True, f=StringIO())
    det.find_signal('a', use_re=True, case_sensitive=True, f=StringIO())
    det.signal_names
    det.report

    cam = det.cam

    cam.image_mode.put('Single')
    # plugins don't live on detectors now:
    # det.image1.enable.put('Enable')
    cam.array_callbacks.put('Enable')

    det.get()
    st = det.trigger()
    repr(st)
    det.read()

    # values = tuple(det.gain_xy.get())
    cam.gain_xy.put(cam.gain_xy.get(), wait=True)

    # fail when only specifying x
    with pytest.raises(ValueError):
        cam.gain_xy.put((0.0, ), wait=True)

    det.describe()
    det.report


def test_tiff_plugin():
    # det = AreaDetector(prefix)
    class TestDet(SimDetector):
        p = Cpt(TIFFPlugin, 'TIFF1:')

    det = TestDet(prefix)
    plugin = det.p

    plugin.file_template.put('%s%s_%3.3d.tif')

    plugin.array_pixels
    plugin


def test_hdf5_plugin():

    class MyDet(SimDetector):
        p = Cpt(HDF5Plugin, suffix='HDF1:')

    d = MyDet(prefix, name='d')
    d.p.file_path.put('/tmp')
    d.p.file_name.put('--')
    d.p.warmup()
    d.stage()
    print(d.p.read_configuration())
    d.p.describe_configuration()
    d.unstage()


def test_subclass():
    class MyDetector(SimDetector):
        tiff1 = Cpt(TIFFPlugin, 'TIFF1:')

    det = MyDetector(prefix)
    det.wait_for_connection()

    print(det.describe())
    print(det.tiff1.capture.describe())


def test_getattr():
    class MyDetector(SimDetector):
        tiff1 = Cpt(TIFFPlugin, 'TIFF1:')

    det = MyDetector(prefix)
    assert getattr(det, 'tiff1.name') == det.tiff1.name
    assert getattr(det, 'tiff1') is det.tiff1
    # raise
    # TODO subclassing issue


def test_invalid_plugins():
    class MyDetector(SingleTrigger, SimDetector):
        tiff1 = Cpt(TIFFPlugin, 'TIFF1:')
        stats1 = Cpt(StatsPlugin, 'Stats1:')

    det = MyDetector(prefix)
    det.wait_for_connection()
    det.tiff1.nd_array_port.put(det.cam.port_name.get())
    det.stats1.nd_array_port.put('AARDVARK')

    with pytest.raises(RuntimeError):
        det.stage()

    with pytest.raises(RuntimeError):
        det.validate_asyn_ports()

    assert ['AARDVARK'] == det.missing_plugins()

def test_validete_plugins_no_portname():
    class MyDetector(SingleTrigger, SimDetector):
        roi1 = Cpt(ROIPlugin, 'ROI1:')
        over1 = Cpt(OverlayPlugin, 'Over1:')

    det = MyDetector(prefix)

    det.roi1.nd_array_port.put(det.cam.port_name.get())
    det.over1.nd_array_port.put(det.roi1.port_name.get())

    det.validate_asyn_ports()

def test_get_plugin_by_asyn_port():
    class MyDetector(SingleTrigger, SimDetector):
        tiff1 = Cpt(TIFFPlugin, 'TIFF1:')
        stats1 = Cpt(StatsPlugin, 'Stats1:')
        roi1 = Cpt(ROIPlugin, 'ROI1:')

    det = MyDetector(prefix)

    det.tiff1.nd_array_port.put(det.cam.port_name.get())
    det.roi1.nd_array_port.put(det.cam.port_name.get())
    det.stats1.nd_array_port.put(det.roi1.port_name.get())

    det.validate_asyn_ports()

    assert det.tiff1 is det.get_plugin_by_asyn_port(det.tiff1.port_name.get())
    assert det.cam is det.get_plugin_by_asyn_port(det.cam.port_name.get())
    assert det.roi1 is det.get_plugin_by_asyn_port(det.roi1.port_name.get())


def test_read_configuration_smoke():
    class MyDetector(SingleTrigger, SimDetector):
        stats1 = Cpt(StatsPlugin, 'Stats1:')
        proc1 = Cpt(ProcessPlugin, 'Proc1:')
        roi1 = Cpt(ROIPlugin, 'ROI1:')

    det = MyDetector(prefix, name='test')
    det.proc1.nd_array_port.put(det.cam.port_name.get())
    det.roi1.nd_array_port.put(det.proc1.port_name.get())
    det.stats1.nd_array_port.put(det.roi1.port_name.get())
    # smoke test
    det.stage()
    conf = det.stats1.read_configuration()
    desc = det.stats1.describe_configuration()
    det.unstage()
    for k in conf:
        assert k in desc

    assert len(conf) > 0
    assert len(conf) == len(desc)


def test_default_configuration_smoke():
    class MyDetector(SimDetector):
        imageplugin = Cpt(ImagePlugin, ImagePlugin._default_suffix)
        statsplugin = Cpt(StatsPlugin, StatsPlugin._default_suffix)
        colorconvplugin = Cpt(ColorConvPlugin, ColorConvPlugin._default_suffix)
        processplugin = Cpt(ProcessPlugin, ProcessPlugin._default_suffix)
        overlayplugin = Cpt(OverlayPlugin, OverlayPlugin._default_suffix)
        roiplugin = Cpt(ROIPlugin, ROIPlugin._default_suffix)
        transformplugin = Cpt(TransformPlugin, TransformPlugin._default_suffix)
        netcdfplugin = Cpt(NetCDFPlugin, NetCDFPlugin._default_suffix)
        tiffplugin = Cpt(TIFFPlugin, TIFFPlugin._default_suffix)
        jpegplugin = Cpt(JPEGPlugin, JPEGPlugin._default_suffix)
        # nexusplugin = Cpt(NexusPlugin, NexusPlugin._default_suffix)
        hdf5plugin = Cpt(HDF5Plugin, HDF5Plugin._default_suffix)
        magickplugin = Cpt(MagickPlugin, MagickPlugin._default_suffix)

    d = MyDetector(prefix, name='d')
    d.stage()
    {n: getattr(d, n).read_configuration() for n in d.signal_names}
    {n: getattr(d, n).describe_configuration() for n in d.signal_names}
    d.unstage()

@pytest.mark.parametrize('plugin',
                         _recursive_subclasses(PluginBase))
def test_default_configuration_attrs(plugin):
    for k in plugin._default_configuration_attrs:
        assert hasattr(plugin, k)
        assert isinstance(getattr(plugin, k),
                          (Component, DynamicDeviceComponent))


from . import main
is_main = (__name__ == '__main__')
main(is_main)



import time
import logging
import unittest

from ophyd import (Device, Component, FormattedComponent)
from ophyd.signal import Signal
from ophyd import EpicsSignal

logger = logging.getLogger(__name__)


class TimeoutTests(unittest.TestCase):
    def test_timeout(self):
        class SubSubDevice(Device):
            cpt5 = Component(EpicsSignal, '5')

        class SubDevice(Device):
            cpt4 = Component(EpicsSignal, '4')
            subsub1 = Component(SubSubDevice, 'sub_')

        class MyDevice(Device):
            sub1 = Component(EpicsSignal, '1')
            sub2 = Component(SubDevice, '_')
            cpt3 = Component(EpicsSignal, '3')

        device = MyDevice('prefix:', name='dev')
        with self.assertRaises(TimeoutError) as cm:
            device.wait_for_connection(timeout=1e-6)

        ex_msg = str(cm.exception)
        self.assertIn('dev.sub1', ex_msg)
        self.assertIn('dev.sub2.cpt4', ex_msg)
        self.assertIn('dev.sub2.subsub1.cpt5', ex_msg)
        self.assertIn('dev.cpt3', ex_msg)

        self.assertIn('prefix:1', ex_msg)
        self.assertIn('prefix:_4', ex_msg)
        self.assertIn('prefix:_sub_5', ex_msg)
        self.assertIn('prefix:3', ex_msg)



import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

from . import *


# Signals
from .signal import (Signal, EpicsSignal, EpicsSignalRO, DerivedSignal)

# Positioners
from .positioner import (PositionerBase, SoftPositioner)
from .epics_motor import EpicsMotor
from .pv_positioner import (PVPositioner, PVPositionerPC)
from .pseudopos import (PseudoPositioner, PseudoSingle)

# Devices
from .scaler import EpicsScaler
from .device import (Device, Component, FormattedComponent,
                     DynamicDeviceComponent)
from .status import StatusBase
from .mca import EpicsMCA, EpicsDXP
from .quadem import QuadEM, NSLS_EM, TetrAMM, APS_EM

# Areadetector-related
from .areadetector import *
from ._version import get_versions

from .commands import (mov, movr, set_pos, wh_pos, set_lm, log_pos,
                       log_pos_diff, log_pos_mov)

from .utils.startup import setup as setup_ophyd


__version__ = get_versions()['version']
del get_versions




"""Command Line Interface to opyd objects"""


import time
import functools
import sys
import warnings
from contextlib import contextmanager, closing
from operator import attrgetter
from io import StringIO
import collections

import IPython
from IPython.utils.coloransi import TermColors as tc

from epics import caget, caput

from . import (EpicsMotor, PositionerBase, PVPositioner, Device)
from .utils import DisconnectedError
from .utils.startup import setup as setup_ophyd
from prettytable import PrettyTable
import numpy as np


__all__ = ['mov',
           'movr',
           'set_pos',
           'wh_pos',
           'set_lm',
           'log_pos',
           'log_pos_diff',
           'log_pos_mov',
           'get_all_positioners',
           'get_logbook',
           'setup_ophyd',
           ]

# Global Defs of certain strings

FMT_LEN = 18
FMT_PREC = 6
DISCONNECTED = 'disconnected'


def scrape_namespace():
    """
    Get all public objects from the user namespace, sorted by name.

    If we are not in an IPython session, warn and return an empty list.
    """
    ip = IPython.get_ipython()
    if ip is None:
        warnings.warn('Unable to inspect Python global namespace; '
                      'use IPython to enable these features.')
        return []
    else:
        return [val for var, val in sorted(ip.user_ns.items())
                if not var.startswith('_')]


def instances_from_namespace(classes):
    '''Get all instances of `classes` from the user namespace

    Parameters
    ----------
    classes : type, or sequence of types
        Passed directly to isinstance(), only instances of these classes
        will be returned.
    '''
    return [val for val in scrape_namespace() if isinstance(val, classes)]


def ducks_from_namespace(attr):
    '''Get all instances that have a given attribute.

    "Ducks" is a reference to "duck-typing." If it looks like a duck....

    Parameters
    ----------
    attr : str
        name of attribute
    '''
    return [val for val in scrape_namespace() if hasattr(val, attr)]


def get_all_positioners():
    '''Get all positioners defined in the IPython namespace'''
    devices = instances_from_namespace((Device, PositionerBase))
    positioners = []
    for device in devices:
        positioners.extend(_recursive_positioner_search(device))
    return positioners


def _recursive_positioner_search(device):
    "Return a flat list the device and any subdevices that can be 'set'."
    # TODO Refactor this as a method on Device.
    res = []

    try:
        if hasattr(device, 'position'):  # duck-typed as a Positioner
            res.append(device)
    except DisconnectedError:
        res.append(device)

    if isinstance(device, Device):  # only Devices have `_signals`
        for d in device._signals.values():
            if isinstance(d, (Device, PositionerBase)):
                res.extend(_recursive_positioner_search(d))
    return res


def _normalize_positioners(positioners):
    "input normalization used by wh_pos, log_pos, log_pos_mov"
    if positioners is None:
        # Grab IPython namespace, recursively find Positioners.
        res = get_all_positioners()
    elif isinstance(positioners, (Device, PositionerBase)):
        # Explore children in case this is a composite Device.
        res = _recursive_positioner_search(positioners)
    else:
        # Assume this is a list of Devices.
        res = []
        for device in positioners:
            if not isinstance(device, (Device, PositionerBase)):
                raise TypeError("Input is not a Device: %r" % device)
            res.extend(_recursive_positioner_search(device))
    return list(sorted(set(res), key=attrgetter('name')))


def var_from_namespace(var):
    ip = IPython.get_ipython()
    if ip is not None:
        return ip.user_ns[var]
    else:
        raise RuntimeError('No IPython session')


def get_logbook():
    '''Get the logbook instance from the user namespace'''
    try:
        return var_from_namespace('logbook')
    except (KeyError, RuntimeError):
        return None


def ensure(*ensure_args):
    def wrap(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            # First check if we have an iterable first on the first arg.
            # If not, then make these all lists
            if len(args) > 0:
                if not hasattr(args[0], "__iter__"):
                    args = tuple([[a] for a in args])
            # Now do type checking ignoring None
            for n, (arg, t) in enumerate(zip(args, ensure_args)):
                if t is None:
                    # Ignore when type is specified as None
                    continue

                invalid = [x for x in arg
                           if not isinstance(x, t)]

                if invalid:
                    raise TypeError('Incorrect type in parameter list.\n'
                                    'Parameter at 0-based position {} must be'
                                    'an instance of {}'.format(n, t))

            f(*args, **kwargs)
        return wrapper
    return wrap


@ensure(PositionerBase, None)
def mov(positioner, position):
    """Move positioners to given positions

    Move positioners using the move method of the Positioner class.

    Parameters
    ----------
    positioner : Positioner or list
        Positioners to move
    position : float or list of float
        Values to move positioners to.

    Examples
    --------
    Move a single positioner `slt1_xc` to 10::

    >>>mov(slt1_xc, 10)

    Move positioner `slt1_xg` and `slt1_yg` to 2 and 3 respectively::

    >>>mov([slt1_xg, slt1_yg], [2, 3])
    """

    print('\n   ', end='')
    print(tc.Green, end='')
    for p in positioner:
        print_string(p.name)
    print("\n")

    # Start Moving all Positioners in context manager to catch
    # Keyboard interrupts

    # TODO : This should be a utility function

    pos_prec = []
    for p in positioner:
        if hasattr(p, 'precision'):
            pos_prec.append(p.precision)
        else:
            pos_prec.append(FMT_PREC)

    with catch_keyboard_interrupt(positioner):
        stat = [p.move(v, wait=False) for p, v in
                zip(positioner, position)]

        # The loop below ensures that at least a couple prints
        # will happen
        flag = 0
        done = False

        while not all(s.done for s in stat) or (flag < 2):
            print(tc.LightGreen, end='')
            print('   ', end='')
            for p, prec in zip(positioner, pos_prec):
                print_value(p.position, egu=p.egu, prec=prec)
            print('\n')
            print('\033[2A', end='')
            time.sleep(0.01)
            done = all(s.done for s in stat)
            if done:
                flag += 1

    print(tc.Normal + '\n')
    for err in [s for s in stat if not s.success]:
        device = err.pos
        reason = "Unknown"
        if isinstance(device, EpicsMotor):
            if device.high_limit_switch.get():
                reason = "Motor reached the high limit switch."
            elif device.low_limit_switch.get():
                reason = "Motor reached the low limit switch."

        print('Warning: {} failed to reach the target position. '
              'Reason: {}'.format(device.name, reason))


@ensure(PositionerBase, None)
def movr(positioner, position):
    """Move positioners relative to their current positon.

    See Also
    --------
    mov : move positioners to an absolute position.
    """

    _start_val = [p.position for p in positioner]
    for v in _start_val:
        if v is None:
            raise IOError("Unable to read motor position for relative move")

    _new_val = [a + b for a, b in zip(_start_val, position)]
    mov(positioner, _new_val)


@ensure(PositionerBase, None)
def set_lm(positioner, limits):
    """Set the limits of the positioner

    Sets the limits of the positioner or list of positioners. For EpicsMotors
    the fields .HLM and .LLM are set to the high and low limits respectively.
    For PVPositioners the .DRVH and .DRVL fields are set on the setopoint
    record. If neither method works then an IOError is raised.

    Parameters
    ----------
    positioner : positioner or list of positioners
    limits : single or list of tuple of form (+ve, -ve) limits

    Raises
    ------
    IOError
        If the caput (EPICS put) fails then an IOError is raised.

    Examples
    --------
    Set the limits of motor `m1` to (10, -10)::

    >>>set_lm(slt1_xc, (10, -10))

    Set the limits of motors `m1` and `m2` to (2, -2) and (3, -3)
    respectively::

    >>>set_lm([m1, m2], [[2,-2], [3, -3]])
    """

    print('')
    msg = ''

    high_fields = []
    low_fields = []
    for p in positioner:
        if isinstance(p, EpicsMotor):
            high_fields.append(p.prefix + '.HLM')
            low_fields.append(p.prefix + '.LLM')
        elif isinstance(p, PVPositioner):
            high_fields.append(p.setpoint_pvname[0] + '.DRVH')
            low_fields.append(p.setpoint_pvname[0] + '.DRVL')
        else:
            raise TypeError("Positioners must be EpicsMotors or PVPositioners"
                            "to set the limits")

    for p, lim, high_field, low_field in zip(positioner,
                                             limits,
                                             high_fields, low_fields):
        lim1 = max(lim)
        lim2 = min(lim)
        if not caput(high_field, lim1):
            raise IOError("Unable to set high limit for {}"
                          " writing to PV {}.".format(p.name, high_field))
        msg += "Upper limit set to {:.{prec}g} for positioner {}\n".format(
               lim1, p.name, prec=FMT_PREC)

        if not caput(low_field, lim2):
            raise IOError("Unable to set low limit for {}"
                          " writing to PV {}.".format(p.name, low_field))
        msg += "Lower limit set to {:.{prec}g} for positioner {}\n".format(
               lim2, p.name, prec=FMT_PREC)

    print(msg)
    logbook = get_logbook()
    if logbook:
        logbook.log(msg)


@ensure(PositionerBase, (float, int))
def set_pos(positioner, position):
    """Set the position of a positioner

    Set the position of a positioner or positioners to the value position.
    This function only works for EpicsMotors (Based on the EPICS Motor Record)
    and uses the .OFF field to set the current position to the value passed to
    the function.

    Parameters
    ----------
    positioner : Positioner or list of positioners.
    position : float or list of floats.
        New position of positioners

    Raises
    ------
    TypeError
        If positioner is not an instance of an EpicsMotor.

    Examples
    --------
    Set the position of motor m1 to 4::

    >>>set_pos(m1, 4)

    Set the position of motors m1 and m2 to 1 and 2 respectively::

    >>>set_pos([m1, m2], [1, 2])

    Raises:
        TypeError: If positioner is not an instance of an EpicsMotor.
    """
    for p in positioner:
        if not isinstance(p, EpicsMotor):
            raise TypeError("Positioner {} must be an EpicsMotor"
                            "to set position.".format(p.name))

    # Get the current offset

    offset_pvs = [p.prefix + ".OFF" for p in positioner]
    dial_pvs = [p.prefix + ".DRBV" for p in positioner]

    old_offsets = [caget(p) for p in offset_pvs]
    dial = [caget(p) for p in dial_pvs]

    for v in old_offsets + dial:
        if v is None:
            raise ValueError("Could not read or invalid value for current"
                             "position of positioners")

    new_offsets = [a - b for a, b in zip(position, dial)]

    msg = ''
    for o, old_o, p in zip(new_offsets, old_offsets, positioner):
        if caput(p.prefix + '.OFF', o):
            msg += 'Motor {0} set to position {1} (Offset = {2} was {3})\n'\
                   .format(p.name, p.position, o, old_o)
        else:
            print('Unable to set position of positioner {0}'.format(p.name))

    print(msg)
    logbook = get_logbook()
    if logbook:
        lmsg = logbook_add_objects(positioner, dial_pvs + offset_pvs)
        logbook.log(msg + '\n' + lmsg)


def wh_pos(positioners=None):
    """Get the current position of Positioners and print to screen.

    Print to the screen the position of the positioners in a formated table.

    Parameters
    ----------
    positioners : Positioner, list of Positioners or None

    See Also
    --------
    log_pos : Log positioner values to logbook

    Examples
    --------
    List all positioners::

    >>>wh_pos()

    List positioners `m1`, `m2` and `m3`::

    >>>wh_pos([m1, m2, m3])
    """
    positioners = _normalize_positioners(positioners)
    _print_pos(positioners, file=sys.stdout)


def log_pos(positioners=None, extra_msg=None):
    """Get the current position of Positioners and make a logbook entry.

    Print to the screen the position of the positioners and make a logbook text
    entry. This routine also creates session information in the logbook so
    positions can be recovered.

    Parameters
    ----------
    positioners : Positioner, list of Positioners or None

    Returns
    -------
    int
        The ID of the logbook entry returned by the logbook.log method.
    """
    positioners = _normalize_positioners(positioners)
    logbook = get_logbook()
    if extra_msg:
        msg = extra_msg + '\n'
    else:
        msg = ''

    with closing(StringIO()) as sio:
        _print_pos(positioners, file=sio)
        msg += sio.getvalue()

    # Add the text representation of the positioners

    # Create the property for storing motor posisions
    pdict = {}
    pdict['values'] = {}

    msg += logbook_add_objects(positioners)

    for p in positioners:
        try:
            pdict['values'][p.name] = p.position
        except DisconnectedError:
            pdict['values'][p.name] = DISCONNECTED

    pdict['objects'] = repr(positioners)
    pdict['values'] = repr(pdict['values'])

    if logbook:
        id_ = logbook.log(msg, properties={'OphydPositioners': pdict},
                          ensure=True)

        print('Logbook positions added as Logbook ID {}'.format(id_))
        return id_


def log_pos_mov(id=None, dry_run=False, positioners=None, **kwargs):
    """Move to positions located in logboook

    This function moves to positions recorded in the experimental logbook using
    the :py:func:`log_pos` function.

    Parameters
    ----------
    id : integer, optional
        ID of logbook entry to search for and move positions to.
    dry_run : bool, optional
        If True, do not move motors, but execute a dry_run
    positioners : list, optional
        List of string names of positioners to compare and move. Other
        positioners in the log entry will be ignored.
    """
    positioners = _normalize_positioners(positioners)
    logpos, objects = logbook_to_objects(id, **kwargs)
    objects = collections.OrderedDict(sorted(objects.items()))

    keys = set(positioners).intersection(set(objects.keys()))
    objects = {x: objects[x] for x in keys}

    print('')
    stat = []
    for key, value in objects.items():
        newpos = logpos[key]
        if newpos == DISCONNECTED:
            print('{}[!!] Unable to move positioner {} {}: position was stored'
                  'as disconnected'.format(tc.Red, key, tc.Normal))
            continue

        try:
            oldpos = value.position
        except DisconnectedError:
            print('{}[!!] Unable to move positioner {} {}: disconnected'
                  ''.format(tc.Red, key, tc.Normal))
            continue

        try:
            if not dry_run:
                stat.append(value.move(newpos, wait=False))
        except Exception as ex:
            print('{}[!!] Unable to move positioner {} {} ({}: {})'
                  ''.format(tc.Red, key, tc.Normal, ex.__class__.__name__, ex))
        else:
            print('{}[**] Moving positioner {} to {}'
                  ' from current position of {}{}`'
                  ''.format(tc.Green, key, newpos, oldpos, tc.Normal))

    print('\n{}Waiting for positioners to complete .....'
          ''.format(tc.LightGreen), end='')

    sys.stdout.flush()

    if len(stat) > 0:
        while all(s.done for s in stat):
            time.sleep(0.01)

    print(' Done{}\n'.format(tc.Normal))


def log_pos_diff(id=None, positioners=None, **kwargs):
    """Move to positions located in logboook

    This function compares positions recorded in the experimental logbook
    using the :py:func:`log_pos` function.

    Parameters
    ----------
    id : integer
        ID of logbook entry to search for and move positions to.
    positioners : list
        List of string names of positioners to compare. Other positioners
        in the log entry will be ignored.
    """

    positioners = _normalize_positioners(positioners)
    logpos, objects = logbook_to_objects(id, **kwargs)
    objects = collections.OrderedDict(sorted(objects.items()))

    # Cycle through positioners and compare position with old value
    # If we have an error, print a warning

    diff = []
    pos = []
    values = []

    keys = set(positioners).intersection(set(objects.keys()))
    objects = {x: objects[x] for x in keys}

    print('')
    for key, value in objects.items():
        oldpos = logpos[key]
        if oldpos == DISCONNECTED:
            print('{}[!!] Unable to compare position {} {}: position was stored'
                  'as disconnected'.format(tc.Red, key, tc.Normal))
            continue

        try:
            newpos = value.position
        except DisconnectedError:
            print('{}[!!] Unable to compare position {} {}: disconnected'
                  ''.format(tc.Red, key, tc.Normal))
            continue

        try:
            diff.append(newpos - oldpos)
        except Exception as ex:
            print('{}[!!] Unable to compare position {}{}: ({}: {})'
                  .format(tc.Red, key, tc.Normal, ex.__class__.__name__, ex))
        else:
            pos.append(value)
            values.append(newpos)

    header_len = 3 * (FMT_LEN + 3) + 1
    print_header(len=header_len)
    print_string('Positioner', pre='| ', post=' | ')
    print_string('Value', post=' | ')
    print_string('Difference', post=' |\n')

    print_header(len=header_len)

    for p, v, d in zip(pos, values, diff):
        print_string(p.name, pre='| ', post=' | ')
        print_value(v, egu=p.egu, post=' | ')
        print_value(d, egu=p.egu, post=' |\n')

    print_header(len=header_len)
    print('')


def logbook_to_objects(id=None, **kwargs):
    """Search the logbook and return positioners"""

    logbook = get_logbook()
    if logbook is None:
        raise RuntimeError("No logbook is available")

    entry = logbook.find(id=id, **kwargs)
    if len(entry) != 1:
        raise ValueError("Search of logbook was not unique, please refine"
                         "search")
    try:
        prop = entry[0]['properties']['OphydPositioners']
    except KeyError:
        raise KeyError('No property in log entry with positioner information')

    try:
        obj = eval(prop['objects'])
        val = eval(prop['values'])
    except Exception as ex:
        raise RuntimeError('Unable to create objects from log entry '
                           '(%s)' % ex)

    objects = {o.name: o for o in obj}
    return val, objects


def logbook_add_objects(objects, extra_pvs=None):
    """Add to the logbook aditional information on ophyd objects.

    This routine takes objects and possible extra pvs and adds to the log entry
    information which is not printed to stdout/stderr.

    Parameters
    ----------
    objects : Ophyd objects
        Objects to add to log entry.
    extra_pvs : List of strings
        Extra PVs to include in report
    """

    msg = ''
    msg += '{:^43}|{:^22}|{:^50}\n'.format('PV Name', 'Name', 'Value')
    msg += '{:-^120}\n'.format('')

    # Make a list of all PVs and positioners
    reports = [o.report for o in objects]
    pvs = [report.get('pv', str(None)) for report in reports]
    names = [o.name for o in objects]
    values = [str(v) for report in reports
              for k, v in report.items() if k != 'pv']

    if extra_pvs is not None:
        pvs += extra_pvs
        names += ['None' for e in extra_pvs]
        values += [caget(e) for e in extra_pvs]

    for a, b, c in zip(pvs, names, values):
        msg += 'PV:{:<40} {:<22} {:<50}\n'.format(a, b, c)

    return msg


def print_header(title='', char='-', len=80, file=sys.stdout):
    print('{:{char}^{len}}'.format(title, char=char, len=len), file=file)


def print_string(val, size=FMT_LEN, pre='', post=' ', file=sys.stdout):
    print('{}{:<{size}}{}'.format(pre, val, post, size=size), end='', file=file)


def print_value(val, prec=FMT_PREC, egu='', **kwargs):
    if val is not None:
        print_string('{: .{fmt}f} {}'.format(val, egu, fmt=prec), **kwargs)
    else:
        print_string('', **kwargs)


def blink(on=True, file=sys.stdout):
    if on:
        print("\x1b[?25h", end='', file=file)
    else:
        print("\x1b[?25l", end='', file=file)


@contextmanager
def catch_keyboard_interrupt(positioners):
    """Context manager to capture Keyboard Interrupt and stop motors

    This context manager should be used when moving positioners via the cli
    to capture the keyboardInterrupt and ensure that motors are stopped and
    clean up the output to the screen.
    """

    blink(False)

    try:
        yield
    except KeyboardInterrupt:
        print(tc.Red + "[!!] ABORTED "
              ": Commanding all positioners to stop.")
        for p in positioners:
            p.stop()
            print("{}[--] Stopping {}{}".format(tc.Red, tc.LightRed, p.name))
        print(tc.Normal, end='')
        blink(True)
        raise
    print(tc.Normal, end='')
    blink(True)


def _print_pos(positioners, file=sys.stdout):
    """Pretty Print the positioners to file"""

    print('', file=file)
    pos = []
    for p in positioners:
        try:
            pos.append(p.position)
        except (DisconnectedError, TypeError):
            pos.append(None)

    # Print out header
    pt = PrettyTable(['Positioner', 'Value', 'Low Limit', 'High Limit'])
    pt.align = 'r'
    pt.align['Positioner'] = 'l'
    pt.float_format = '8.5'

    for p, v in zip(positioners, pos):
        if pos is None:
            continue
        if v is not None:
            try:
                prec = p.precision
            except (AttributeError, DisconnectedError):
                prec = FMT_PREC
            value = np.round(v, decimals=prec)
        else:
            value = DISCONNECTED

        try:
            low_limit, high_limit = p.low_limit, p.high_limit
        except DisconnectedError:
            low_limit = high_limit = DISCONNECTED

        pt.add_row([p.name, value, low_limit, high_limit])

    print(pt, file=file)



#!/usr/bin/env python
from setuptools import (setup, find_packages)
import versioneer


setup(name='ophyd',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      license='BSD',
      packages=find_packages())



import logging

import atexit
import epics
import time

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__all__ = ['setup', 'logger']

_dispatcher = None


def setup():
    '''Setup ophyd for use

    Must be called once per session using ophyd
    '''
    # It's important to use the same context in the callback dispatcher
    # as the main thread, otherwise not-so-savvy users will be very
    # confused
    global _dispatcher

    if _dispatcher is not None:
        logger.debug('ophyd already setup')
        return

    epics.ca.use_initial_context()

    from .epics_pvs import MonitorDispatcher
    logger.debug('Installing monitor dispatcher')
    _dispatcher = MonitorDispatcher()
    atexit.register(_cleanup)
    return _dispatcher


def _cleanup():
    '''Clean up the ophyd session'''
    global _dispatcher
    if _dispatcher is None:
        return

    logger.debug('Performing ophyd cleanup')
    if _dispatcher.is_alive():
        logger.debug('Joining the dispatcher thread')
        _dispatcher.stop()
        _dispatcher.join()

    _dispatcher = None

    logger.debug('Finalizing libca')
    epics.ca.finalize_libca()



from ophyd.status import StatusBase


def _setup_st():
    st = StatusBase()
    state = {}

    def cb():
        state['done'] = True
    repr(st)
    return st, state, cb



def test_status_post():
    st, state, cb = _setup_st()

    assert 'done' not in state
    st.finished_cb = cb
    assert 'done' not in state
    st._finished()
    assert 'done' in state
    assert state['done']


def test_status_pre():
    st, state, cb = _setup_st()

    st._finished()

    assert 'done' not in state
    st.finished_cb = cb
    assert 'done' in state
    assert state['done']
